extern crate chrono;
extern crate feed_rs;
extern crate maud;
extern crate reqwest;

use rand::seq::SliceRandom;
use rand::thread_rng;
use std::error::Error;
use std::fs::write;
use std::fs::DirBuilder;
use std::path::Path;
mod site_generator;
mod utilities;
mod web_fetchers;
use rayon::prelude::*;

fn main() -> Result<(), Box<dyn Error>> {
    setup_logging()?;

    let posts = load_and_process_posts()?;
    let archive_posts = create_archive_posts(&posts)?;

    generate_and_write_output(posts, archive_posts)?;

    Ok(())
}

fn setup_logging() -> Result<(), Box<dyn Error>> {
    simple_logger::init_with_level(log::Level::Info)?;
    Ok(())
}

fn load_and_process_posts() -> Result<Vec<utilities::Post>, Box<dyn Error>> {
    let mut posts = grab_posts()?;
    process_posts(&mut posts)?;
    Ok(posts)
}

fn grab_posts() -> Result<Vec<utilities::Post>, Box<dyn Error>> {
    let mut posts = utilities::read_feed("feeds.txt");
    posts.retain(|post| post.score.is_positive());
    utilities::retain_most_recent_based_on_main_url(&mut posts);

    let mut featured = utilities::read_feed("featured_feeds.txt");
    featured.iter_mut().for_each(|post| {
        post.score = (post.score as f64 * 1.5) as i64;
    });
    posts.extend(featured);

    Ok(posts)
}

fn process_posts(posts: &mut Vec<utilities::Post>) -> Result<(), Box<dyn Error>> {
    posts.par_iter_mut().for_each(|post| {
        utilities::find_image(post);
        utilities::validate(post);
    });

    posts.sort();

    if let Some(pos) = posts.iter().position(|post| post.image_url.is_some()) {
        let post_with_image = posts.remove(pos);
        posts.insert(0, post_with_image);
    }

    posts.truncate(16);
    Ok(())
}

fn create_archive_posts(posts: &[utilities::Post]) -> Result<Vec<utilities::Post>, Box<dyn Error>> {
    const ARCHIVE_SIZE: usize = 100;

    let mut old_posts: Vec<_> = utilities::read_feed("feeds.txt")
        .into_iter()
        .filter(|p| !posts.contains(p))
        .collect();
    old_posts.shuffle(&mut thread_rng());

    let mut archive_posts = Vec::new();
    while archive_posts.len() < ARCHIVE_SIZE && !old_posts.is_empty() {
        let chunk_size = std::cmp::min(ARCHIVE_SIZE - archive_posts.len() + 50, old_posts.len());
        let mut chunk: Vec<_> = old_posts.drain(..chunk_size).collect();

        chunk.par_iter_mut().for_each(utilities::validate);
        chunk.retain(|post| post.score != 0);
        archive_posts.extend(chunk);
    }

    archive_posts.truncate(ARCHIVE_SIZE);
    Ok(archive_posts)
}

fn generate_and_write_output(
    posts: Vec<utilities::Post>,
    archive_posts: Vec<utilities::Post>,
) -> Result<(), Box<dyn Error>> {
    let index = site_generator::generate_index(posts.clone(), archive_posts);
    write_to_file("output/index.html", index.into_string())?;

    let feed = site_generator::generate_rss(posts);
    write_to_file("output/feed.xml", feed)?;

    Ok(())
}

fn write_to_file<P: AsRef<Path>>(path: P, content: String) -> Result<(), Box<dyn Error>> {
    let path = path.as_ref();
    DirBuilder::new()
        .recursive(true)
        .create(path.parent().unwrap())?;
    write(path, content)?;
    log::info!("Successfully wrote to {}", path.display());
    Ok(())
}
